import logging
import pathlib

from qtpy import QtWidgets

from ..runtime.appdata import AppData
from ..utils import FileBrowser
from ..io.catalogue import load_cat, create_cat
from ..io.viewer_data import get_ids_from_dir
from ..utils.logs import qlog


logger = logging.getLogger(__name__)


class NewFile(QtWidgets.QDialog):
    def __init__(self, rd: AppData, parent=None):
        self.rd = rd

        super().__init__(parent)
        self.setWindowTitle("Create a New Inspection File")

        layout = QtWidgets.QVBoxLayout()
        layout.setSpacing(20)

        self.setFixedSize(layout.sizeHint())

        self._browsers = {
            'output': FileBrowser(title='Output File:', filename_extensions='CSV Files (*.csv)',
                                  mode=FileBrowser.SaveFile, default_path=pathlib.Path().resolve() / 'Untitled.csv',
                                  parent=self),
            'data': FileBrowser(title='Data Source:', mode=FileBrowser.OpenDirectory,
                                default_path=self.rd.config.data.dir, parent=self),
            'cat': FileBrowser(title='Catalogue:', filename_extensions='FITS Files (*.fits)', mode=FileBrowser.OpenFile,
                               default_path=self.rd.config.cat.filename, parent=self)
        }

        # add a file browser for specifying the output file
        layout.addWidget(self._browsers['output'])

        # add a file browser for specifying the data directory
        layout.addWidget(self._browsers['data'])

        # add radio buttons for choosing between creating a new catalogue and loading an existing one
        self._cat_factory_group = QtWidgets.QButtonGroup()
        self._cat_factory_group.buttonToggled.connect(self.update_cat_factory_state)

        self._create_cat_radio_button = QtWidgets.QRadioButton('Create a new catalogue')
        self._cat_factory_group.addButton(self._create_cat_radio_button)

        self._load_cat_radio_button = QtWidgets.QRadioButton('Load an existing catalogue')
        self._cat_factory_group.addButton(self._load_cat_radio_button)

        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.addWidget(self._create_cat_radio_button)
        sub_layout.addWidget(self._load_cat_radio_button)
        layout.addLayout(sub_layout)

        # add a horizontal separator
        self.separator = QtWidgets.QFrame()
        self.separator.setFrameShape(QtWidgets.QFrame.HLine)
        layout.addWidget(self.separator)

        # add a file browser for specifying the catalogue
        layout.addWidget(self._browsers['cat'])
        self._browsers['cat'].setHidden(True)

        # add a checkbox for specifying the catalogue loader mode
        self._filter_check_box = QtWidgets.QCheckBox(
            'Filter the catalogue using a list of IDs extracted from the data directory')
        self._filter_check_box.setChecked(False)
        self._filter_check_box.setHidden(True)
        self._filter_check_box.stateChanged.connect(self.update_id_pattern_widget_state)
        layout.addWidget(self._filter_check_box)

        # add a line edit for specifying the ID pattern that will be used to parse the filenames in the data directory
        self._id_pattern_label = QtWidgets.QLabel('ID pattern:')
        self._id_pattern_label.setToolTip('A regular expression used to match the source IDs while scanning '
                                          'the data directory. If more than\none ID is matched to the pattern for a '
                                          'given filename, only the longest match will be returned.')
        self._id_pattern_label.setFixedWidth(120)
        self._id_pattern_label.setSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Minimum)

        self._id_pattern = QtWidgets.QLineEdit(self.rd.config.data.id_pattern)

        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.addWidget(self._id_pattern_label)
        sub_layout.addWidget(self._id_pattern)
        layout.addLayout(sub_layout)

        # add OK/Cancel buttons
        self._button_box = QtWidgets.QDialogButtonBox(QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel)
        layout.addWidget(self._button_box)
        self._button_box.accepted.connect(self.accept)
        self._button_box.rejected.connect(self.reject)

        # set the state of the radio buttons
        self._create_cat_radio_button.setChecked(True)

        self.setLayout(layout)

    @qlog
    def get_catalogue(self):
        # validate the input
        for b in self._browsers.values():
            if not b.isHidden() and (not b.is_filled() or not b.exists()):
                return

        # create a new catalogue if necessary
        if self._browsers['cat'].isHidden():
            ids = get_ids_from_dir(self._browsers['data'].path, self._id_pattern.text())
            if ids is None:
                return
            return create_cat(ids)

        # otherwise, load an existing catalogue
        translate = self.rd.config.cat.translate
        data_dir = self._browsers['data'].path if self._filter_check_box.isChecked() else None

        return load_cat(self._browsers['cat'].path, translate=translate, data_dir=data_dir,
                        id_pattern=self.rd.config.data.id_pattern)

    def accept(self):
        cat = self.get_catalogue()
        if cat is None:
            return

        self.rd.cat = cat
        self.rd.output_path = pathlib.Path(self._browsers['output'].path)

        # update the user configuration file
        self.rd.config.data.dir = self._browsers['data'].path
        if not self._browsers['cat'].isHidden():
            self.rd.config.cat.filename = self._browsers['cat'].path
        if not self._id_pattern.isHidden():
            self.rd.config.data.id_pattern = self._id_pattern.text()
        self.rd.config.save(self.rd.config_file)

        self.rd.create()

        super().accept()

    def update_cat_factory_state(self):
        if self._create_cat_radio_button.isChecked():
            self._browsers['cat'].setHidden(True)
            self._filter_check_box.setHidden(True)

            self._id_pattern.setHidden(False)
            self._id_pattern_label.setHidden(False)
        else:
            self._browsers['cat'].setHidden(False)
            self._filter_check_box.setHidden(False)

            self.update_id_pattern_widget_state()

    def update_id_pattern_widget_state(self):
        if self._filter_check_box.isChecked():
            self._id_pattern.setHidden(False)
            self._id_pattern_label.setHidden(False)
        else:
            self._id_pattern.setHidden(True)
            self._id_pattern_label.setHidden(True)



import logging
from dataclasses import asdict

import numpy as np
from astropy.utils.decorators import lazyproperty

import pyqtgraph as pg
from qtpy import QtCore, QtWidgets

from ..utils.params import read_yaml
from ..utils import SmartSlider
from ..utils import tables

from .ViewerElement import ViewerElement
from ..runtime.appdata import AppData
from ..runtime import config


logger = logging.getLogger(__name__)


class Spec1D(ViewerElement):
    def __init__(self, rd: AppData, cfg: config.Spectrum, alias: str, parent=None):
        super().__init__(rd=rd, cfg=cfg, alias=alias, parent=parent)

        self.cfg = cfg

        # load the list of spectral lines
        # TODO: move to the application data
        self._lines = read_yaml('default_lines.yml', in_dist=True)

        # create a widget for the spectrum
        self._spec_1d_widget = pg.GraphicsView()
        self._spec_1d_layout = pg.GraphicsLayout()
        self._spec_1d_widget.setCentralItem(self._spec_1d_layout)

        # create a redshift slider
        self._z_slider = SmartSlider(QtCore.Qt.Horizontal, **asdict(self.cfg.slider))
        self._z_slider.valueChanged[int].connect(self._update_from_slider)
        self._z_slider.setToolTip('Slide to redshift.')

        # create a line edit for changing the redshift
        self._redshift_editor = QtWidgets.QLineEdit()
        self._redshift_editor.returnPressed.connect(self._update_from_editor)
        self._redshift_editor.setMaximumWidth(120)

        # set up the plot
        self._spec_1d = self._spec_1d_layout.addPlot(name=alias)
        self._spec_1d.setMouseEnabled(True, True)
        self._label_style = {'color': 'r', 'font-size': '20px'}

        # set up the spectral lines
        self._line_artists = {}
        # TODO: store colors in config
        line_color = (175.68072, 220.68924, 46.59488)
        line_pen = pg.mkPen(color=line_color, width=1)
        for line_name, lambda0 in self._lines['lambda'].items():
            line = pg.InfiniteLine(angle=90, movable=False, pen=line_pen)
            label = pg.TextItem(text=line_name, color=line_color, anchor=(1, 1), angle=-90)

            self._line_artists[line_name] = {'line': line, 'label': label}

    def init_ui(self):
        self.layout.addWidget(self._spec_1d_widget, 1, 1, 1, 3)
        self.layout.addWidget(self._z_slider, 2, 1, 1, 1)
        self.layout.addWidget(QtWidgets.QLabel('z = ', self), 2, 2, 1, 1)
        self.layout.addWidget(self._redshift_editor, 2, 3, 1, 1)

    @lazyproperty
    def default_xrange(self):
        return np.nanmin(self.data['wavelength']), np.nanmax(self.data['wavelength'])

    @lazyproperty
    def default_yrange(self):
        return np.nanmin(self.data['flux']), np.nanmax(self.data['flux'])

    @lazyproperty
    def _label_height(self):
        y_min, y_max = self.default_yrange
        return y_min + 0.6 * (y_max - y_min)

    def _plot(self):
        self._spec_1d.plot(self.data['wavelength'], self.data['flux'], pen='k')

        if 'flux_error' in self.data.colnames:
            self._spec_1d.plot(self.data['wavelength'], self.data['flux_error'], pen='r')

        for line_name, line_artist in self._line_artists.items():
            self._spec_1d.addItem(line_artist['line'], ignoreBounds=True)
            self._spec_1d.addItem(line_artist['label'])

        for keyword, position in {'TUNIT1': 'bottom', 'TUNIT2': 'right'}.items():
            if self.meta.get(keyword):
                self._spec_1d.setLabel(position, self.meta[keyword], **self._label_style)

    def _update_view(self):
        for line_name, line_artist in self._line_artists.items():
            line_wave = self._lines['lambda'][line_name] * (1 + self._z_slider.value)
            line_artist['line'].setPos(line_wave)
            line_artist['label'].setPos(QtCore.QPointF(line_wave, self._label_height))

    def _update_from_slider(self, index=None):
        if index is not None:
            self._z_slider.index = index
        self._redshift_editor.setText("{:.6f}".format(self._z_slider.value))
        self._update_view()

    def _update_from_editor(self):
        try:
            self._z_slider.index = self._z_slider.index_from_value(float(self._redshift_editor.text()))
        except ValueError:
            logger.error('Invalid redshift value: {}'.format(self._redshift_editor.text()))
            self._z_slider.reset()

        self._update_from_slider()

    def load_object(self):
        super().load_object()
        if self.data is None:
            return

        for cname in ('wavelength', 'flux'):
            if cname not in self.data.colnames:
                logger.error(tables.column_not_found_message(cname, self.rd.config.data.translate))
                return

    def display(self):
        try:
            self._z_slider.default_value = self.rd.cat.loc[self.rd.id]['z']
        except KeyError:
            self._z_slider.default_value = self.cfg.slider.default_value

        self._plot()

    def reset_view(self):
        if self.data is None:
            return

        self._z_slider.reset()
        self._update_from_slider()

        self._spec_1d.setXRange(*self.default_xrange)
        self._spec_1d.setYRange(*self.default_yrange)

    def clear_content(self):
        del self.default_xrange
        del self.default_yrange
        del self._label_height

        self._redshift_editor.setText("")
        self._spec_1d.clear()



from dataclasses import dataclass

from ..utils.params import Params, LocalFile


@dataclass
class Cat:
    filename: str | None
    translate: dict[str, list[str]] | None


@dataclass
class Data:
    dir: str
    id_pattern: str
    translate: dict[str, list[str]] | None


@dataclass
class Appearance:
    antialiasing: bool = False


@dataclass
class AbstractWidget:
    pass


@dataclass
class ControlPanel(AbstractWidget):
    button_width: int


@dataclass
class ObjectInfo(AbstractWidget):
    items: dict[str, str] | None


@dataclass
class ReviewForm(AbstractWidget):
    checkboxes: dict[str, str] | None


@dataclass
class ViewerElement(AbstractWidget):
    filename_keyword: str
    loader: str | None
    loader_config: dict[str, str] | None
    position: str | None
    relative_to: str | None


@dataclass
class Image(ViewerElement):
    rotate: int | None
    scale: float | None
    interactive: bool = True


@dataclass
class Slider:
    min_value: float | None
    max_value: float | None
    step: float | None
    default_value: float | None


@dataclass
class Spectrum(ViewerElement):
    slider: Slider


@dataclass
class Viewer(AbstractWidget):
    images: dict[str, Image] | None
    spectra: dict[str, Spectrum] | None


@dataclass
class Config(Params):
    cat: Cat
    data: Data
    appearance: Appearance
    control_panel: ControlPanel
    object_info: ObjectInfo
    review_form: ReviewForm
    viewer: Viewer
    plugins: list[str]



import importlib

from pyqtgraph.dockarea.Dock import Dock
from pyqtgraph.dockarea.DockArea import DockArea

from .AbstractWidget import AbstractWidget
from .ViewerElement import ViewerElement
from .Image2D import Image2D
from .Spec1D import Spec1D

from ..runtime.appdata import AppData
from ..runtime import config


class DataViewer(AbstractWidget):
    def __init__(self, rd: AppData, cfg: config.Viewer, plugins=None, parent=None):
        super().__init__(cfg=cfg, parent=parent)

        self.rd = rd
        self.cfg = cfg

        if plugins is not None:
            self._plugins = [importlib.import_module("specvizitor.plugins." + plugin_name).Plugin() for plugin_name in plugins]
        else:
            self._plugins = []

        self.layout.setSpacing(10)
        self.layout.setContentsMargins(0, 0, 0, 0)

        self.dock_area = DockArea()
        self.dock_widgets = self.create_widgets()
        self.docks = self.create_docks()

    def init_ui(self):
        self.layout.addWidget(self.dock_area, 1, 1, 1, 1)
        self.add_docks()

        try:
            # set `extra` to None to catch an exception (KeyError) when adding extra docks not mentioned in `state`
            self.dock_area.restoreState(self.rd.cache.dock_state, missing='ignore', extra=None)
        except (KeyError, ValueError, TypeError):
            for d in self.docks.values():
                d.close()
            self.docks = self.create_docks()
            self.add_docks()

        for w in self.dock_widgets.values():
            w.init_ui()

    def create_widgets(self) -> dict[str, ViewerElement]:
        widgets = {}

        # create widgets for images (e.g. image cutouts, 2D spectra)
        if self.cfg.images is not None:
            for name, image_cfg in self.cfg.images.items():
                widgets[name] = Image2D(rd=self.rd, cfg=image_cfg, alias=name, parent=self)

        # create widgets for 1D spectra
        if self.cfg.spectra is not None:
            for name, spec_cfg in self.cfg.spectra.items():
                widgets[name] = Spec1D(rd=self.rd, cfg=spec_cfg, alias=name, parent=self)

        return widgets

    def create_docks(self) -> dict[str, Dock]:
        docks = {}
        for name, widget in self.dock_widgets.items():
            dock = Dock(name)
            dock.addWidget(widget)
            docks[name] = dock
        return docks

    def add_docks(self):
        for name, d in self.docks.items():
            position = self.dock_widgets[name].cfg.position

            self.dock_area.addDock(dock=d,
                                   position=position if position is not None else 'bottom',
                                   relativeTo=self.dock_widgets[name].cfg.relative_to)

    def save_dock_state(self):
        self.rd.cache.dock_state = self.dock_area.saveState()
        self.rd.cache.save(self.rd.cache_file)

    def load_object(self):
        self.save_dock_state()

        for name, w in self.dock_widgets.items():
            # clear the widget content
            w.clear_content()

            # load the data to the widget
            w.load_object()

            # display the data
            if w.data is not None:
                w.setEnabled(True)
                w.display()
                w.reset_view()
            else:
                w.setEnabled(False)

            # update the title of the dock
            if w.filename is not None and w.data is not None:
                self.docks[name].setTitle(w.filename.name)
            else:
                self.docks[name].setTitle(name)

        for plugin in self._plugins:
            plugin.link(self.dock_widgets)

    def reset_view(self):
        for w in self.dock_widgets.values():
            w.reset_view()

    def take_screenshot(self, filename: str):
        self.grab().save(filename)



import enum
import logging
import pathlib
import re
from enum import Enum

from astropy.io import fits
from astropy.table import Table
from PIL import Image, ImageOps
import numpy as np

logger = logging.getLogger(__name__)


def load_fits(filename: pathlib.Path, extname: str = None, extver: str = None):

    hdul = fits.open(filename)

    if extname is not None and extver is None:
        index = extname
    elif extname is not None and extver is not None:
        index = (extname, extver)
    else:
        index = 1

    try:
        hdu = hdul[index]
    except KeyError:
        logger.error(f'Extension `{index}` not found (filename: {filename.name})')
        return None, None

    data = hdu.data
    meta = hdu.header

    if meta['XTENSION'] in ('TABLE', 'BINTABLE'):
        data = Table(data)

    return data, meta


def load_pil(filename: pathlib.Path):
    image = Image.open(filename)
    image = ImageOps.flip(image)
    return np.asarray(image), image.info


def load(loader_name: str | None, filename: pathlib.Path, alias: str, **kwargs):
    loaders = {
        'fits': load_fits,
        'pil': load_pil
    }

    if loader_name is None:
        if filename.suffix == '.fits':
            loader_name = 'fits'
        else:
            loader_name = 'pil'

    if loader_name not in loaders.keys():
        logger.error(f'Unknown loader type: `{loader_name}` ({alias}). Available loaders: {tuple(loaders.keys())}')
        return None, None

    return loaders[loader_name](filename, **kwargs)


def get_filename(directory, pattern: str, object_id) -> pathlib.Path | None:
    """ Find a file for a given pattern and a given object ID. If more than one file name is matched to the pattern,
    return the first item from an alphabetically ordered list of matched file names.

    @param directory: the directory where the search is performed
    @param pattern: the pattern used to match file names
    @param object_id: the object ID
    @return: the file name matched to the pattern
    """

    filenames = sorted(pathlib.Path(directory).glob('*'))

    # match to the pattern
    matched_to_pattern = [p for p in filenames if re.search(pattern, str(p))]

    # match to the ID
    if isinstance(object_id, int):
        # make sure that we don't match e.g. '1123' or '1234' to '123' (but match '0123')
        matched_ids = [re.findall(r'\d*{}\d*'.format(object_id), str(p)) for p in matched_to_pattern]
        matched = [p for i, p in enumerate(matched_to_pattern)
                   if matched_ids[i] and str(max(matched_ids[i], key=len)).lstrip('0') == str(object_id)]
    else:
        matched = [p for p in matched_to_pattern if re.search(f'{object_id}', str(p))]

    if matched:
        return matched[0]
    else:
        return


def get_id_from_filename(filename, pattern: str) -> str | None:
    """ Extract the object ID from a file name using a pattern. If more than one ID is matched to the pattern, return
    the longest match (a typical case for integer IDs).

    @param filename: the file name to parse
    @param pattern: the pattern (a regular expression) used to match the ID
    @return: the matched ID
    """

    try:
        matches: list[str] = re.findall(pattern, pathlib.Path(filename).name)
    except re.error:
        return

    if matches:
        return max(matches, key=len)
    else:
        return


def get_ids_from_dir(directory, id_pattern: str) -> np.ndarray | None:
    """ Extract IDs from a directory using a pattern.
    @param directory: the directory where the search for IDs is performed
    @param id_pattern: the pattern (a regular expression) used to match IDs
    @return: a list of matched IDs
    """

    data_files = sorted(pathlib.Path(directory).glob('*'))  # includes subdirectories, if any
    ids = [get_id_from_filename(p, id_pattern) for p in data_files]
    ids = np.array([i for i in ids if i is not None])

    try:
        # convert IDs to int
        ids = ids.astype(np.int64)
        logger.info('Converted IDs to int')
    except ValueError:
        pass

    # remove ID duplicates
    ids = np.unique(ids)

    if not ids.size:
        logger.error('No IDs retrieved from the data directory')
        return

    return ids



from astropy.table import Table


def translate(t: Table, dictionary: dict):
    for cname, cname_synonyms in dictionary.items():
        for syn in cname_synonyms:
            if syn in t.colnames:
                t.rename_column(syn, cname)
                break


def column_not_found_message(cname: str, dictionary: dict[str, list[str]] | None = None) -> str:
    """ Create the `column not found` message.
    @param cname: the column name
    @param dictionary:
    @return: the message
    """

    if dictionary is None or cname not in dictionary:
        return '`{}` column not found'.format(cname)
    else:
        return '`{}` column and its equivalences ({}) not found'.format(cname, ", ".join(dictionary[cname]))



import logging

import numpy as np
import pyqtgraph as pg
from astropy.visualization import ZScaleInterval

from qtpy import QtWidgets
from pgcolorbar.colorlegend import ColorLegendItem

from .ViewerElement import ViewerElement
from ..runtime.appdata import AppData
from ..runtime import config


logger = logging.getLogger(__name__)


class Image2D(ViewerElement):
    def __init__(self, rd: AppData, cfg: config.Image, alias: str, parent=None):
        super().__init__(rd=rd, cfg=cfg, alias=alias, parent=parent)

        self.cfg = cfg

        # add a widget for the image
        self._image_2d_widget = pg.GraphicsView()
        self._image_2d_layout = pg.GraphicsLayout()
        self._image_2d_widget.setCentralItem(self._image_2d_layout)

        if self.cfg.interactive:
            # set up the color map
            self._cmap = pg.colormap.get('viridis')

            # set up the image and the view box
            self.image_2d_plot = self._image_2d_layout.addPlot(name=alias)

            self.image_2d = pg.ImageItem(border='k')
            self.image_2d.setLookupTable(self._cmap.getLookupTable())
            self.image_2d_plot.addItem(self.image_2d)

            # set up the color bar
            self._cbar = ColorLegendItem(imageItem=self.image_2d, showHistogram=True, histHeightPercentile=99.0)
            self._image_2d_layout.addItem(self._cbar, 0, 1)

            # lock the aspect ratio
            self.image_2d_plot.setAspectLocked(True)

        else:
            self._view_box = pg.ViewBox()
            self._image_2d_layout.setContentsMargins(0, 0, 0, 0)
            self._view_box.setAspectLocked(True)
            self.image_2d = pg.ImageItem()
            self._view_box.addItem(self.image_2d)
            self._image_2d_layout.addItem(self._view_box)

    def init_ui(self):
        self.layout.addWidget(self._image_2d_widget, 1, 1)

    def load_object(self):
        super().load_object()
        if self.data is None:
            return

        # rotate the image
        if self.cfg.rotate is not None:
            self.data = np.rot90(self.data, k=self.cfg.rotate // 90)

        # scale the data points
        if self.cfg.scale is not None:
            self.data *= self.cfg.scale

    def display(self):
        self.image_2d.setImage(self.data)

    def reset_view(self):
        if self.data is None:
            return

        if self.cfg.interactive:
            # TODO: allow to choose between min/max and zscale?
            self._cbar.setLevels(ZScaleInterval().get_limits(self.data))
            self.image_2d_plot.autoRange()
        else:
            self._view_box.autoRange(padding=0)

    def clear_content(self):
        self.image_2d.clear()



import logging

import numpy as np
from astropy.io import fits
from astropy.table import Table

from .viewer_data import get_ids_from_dir
from ..utils import tables

logger = logging.getLogger(__name__)


def load_cat(filename=None,
             translate: dict[str, list[str]] | None = None,
             data_dir=None,
             id_pattern=r'\d+') -> [Table, None]:

    """ Read and filter the catalogue.
    @param filename: the catalogue filename
    @param translate:
    @param data_dir:
    @param id_pattern:
    @return: the processed catalogue
    """

    if filename is None:
        logger.warning('Catalogue filename not specified')
        return

    # load the catalogue
    try:
        cat = Table(fits.getdata(filename))
    except OSError:
        logger.error('Failed to load the catalogue')
        return

    # rename columns
    if translate is not None:
        tables.translate(cat, translate)

    # check that the ID column is present in the catalogue
    if 'id' not in cat.colnames:
        logger.error(tables.column_not_found_message('id', translate))
        return
    cat.add_index('id')

    if data_dir is not None:
        ids = get_ids_from_dir(data_dir, id_pattern)
        if ids is None:
            return

        # filter objects based on the list of IDs
        cat = cat[np.in1d(cat['id'], ids, assume_unique=True)]

    if len(cat) == 0:
        logger.error('The processed catalogue is empty')
        return

    return cat


def create_cat(ids) -> Table:
    cat = Table([ids], names=('id',))
    cat.add_index('id')

    return cat



import logging
import pathlib
from functools import partial

import numpy as np

import qtpy
from qtpy import QtGui, QtCore, QtWidgets

from ..runtime.appdata import AppData
from ..runtime import config
from .AbstractWidget import AbstractWidget


logger = logging.getLogger(__name__)


class ControlPanel(QtWidgets.QGroupBox, AbstractWidget):
    reset_button_clicked = QtCore.Signal()
    screenshot_button_clicked = QtCore.Signal(str)
    object_selected = QtCore.Signal(int)

    def __init__(self, rd: AppData, cfg: config.ControlPanel, parent=None):
        super().__init__(cfg=cfg, parent=parent)

        self.rd = rd
        self.cfg = cfg

        self.setTitle('Control Panel')
        self.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        # create the ID button
        self._id_button = QtWidgets.QPushButton()
        self._id_button.setText('ID --')
        self._id_button.setToolTip('Reset the view')
        self._id_button.setFixedWidth(self.cfg.button_width)
        self._id_button.clicked.connect(self.reset_button_clicked.emit)

        # create a widget displaying the index of the current object and the total number of objects in the catalogue
        self._number_of_obj_label = QtWidgets.QLabel()

        # create buttons for switching to the next or previous object
        self._pn_buttons = self.create_pn_buttons()

        # create the `starred` button
        self._star_button = QtWidgets.QPushButton()
        self._star_button.setIcon(QtGui.QIcon(self.get_star_icon()))
        self._star_button.setToolTip('Star the object')
        self._star_button.clicked.connect(self.star)

        # create the `screenshot` button
        self._screenshot_button = QtWidgets.QPushButton()
        self._screenshot_button.setIcon(QtGui.QIcon(get_icon_abs_path('screenshot.svg')))
        self._screenshot_button.setToolTip('Take a screenshot')
        self._screenshot_button.clicked.connect(self.screenshot)

        # create the `reset view` button
        self._reset_view_button = QtWidgets.QPushButton()
        self._reset_view_button.setIcon(QtGui.QIcon(get_icon_abs_path('reset-view.svg')))
        self._reset_view_button.setToolTip('Reset the view')
        self._reset_view_button.clicked.connect(self.reset_button_clicked.emit)

        # create a `dark mode` button
        self._dark_mode = QtWidgets.QPushButton()
        self._dark_mode.setIcon(QtGui.QIcon(get_icon_abs_path('dark-mode.svg')))
        self._dark_mode.setToolTip('Turn on the dark theme')

        # create the `Go to ID` button
        self._go_to_id_button = QtWidgets.QPushButton()
        self._go_to_id_button.setText('Go to ID')
        self._go_to_id_button.setFixedWidth(self.cfg.button_width)
        self._go_to_id_button.clicked.connect(self.go_to_id)

        self._id_field = QtWidgets.QLineEdit()
        self._id_field.setFixedWidth(self.cfg.button_width)
        self._id_field.returnPressed.connect(self.go_to_id)

        # create the `Go to index` button
        self._go_to_index_button = QtWidgets.QPushButton()
        self._go_to_index_button.setText('Go to #')
        self._go_to_index_button.setFixedWidth(self.cfg.button_width)
        self._go_to_index_button.clicked.connect(self.go_to_index)

        self._index_field = QtWidgets.QLineEdit()
        self._index_field.setFixedWidth(self.cfg.button_width)
        self._index_field.returnPressed.connect(self.go_to_index)

    def create_pn_buttons(self) -> dict[str, QtWidgets.QPushButton]:
        pn_buttons_params = {'previous': {'shortcut': 'left', 'icon': 'arrow-backward'},
                             'next': {'shortcut': 'right', 'icon': 'arrow-forward'},
                             'previous starred': {'icon': 'arrow-backward-starred'},
                             'next starred': {'icon': 'arrow-forward-starred'}
                             }

        pn_buttons = {}
        for pn_text, pn_properties in pn_buttons_params.items():
            button = QtWidgets.QPushButton('')
            button.setToolTip('Go to the {} object'.format(pn_text))
            button.setIcon(QtGui.QIcon(get_icon_abs_path(pn_properties['icon'])))

            button.clicked.connect(partial(self.previous_next_object, pn_text.split(' ')[0], 'starred' in pn_text))

            if pn_properties.get('shortcut'):
                button.setShortcut(pn_properties['shortcut'])

            pn_buttons[pn_text] = button

        return pn_buttons

    def init_ui(self):
        self.layout.addWidget(self._id_button, 1, 1, 1, 2)
        self.layout.addWidget(self._number_of_obj_label, 1, 3, 1, 2)

        self.layout.addWidget(self._pn_buttons['previous'], 2, 1, 1, 1)
        self.layout.addWidget(self._pn_buttons['next'], 2, 2, 1, 1)
        self.layout.addWidget(self._star_button, 2, 3, 1, 1)
        self.layout.addWidget(self._dark_mode, 2, 4, 1, 1)

        self.layout.addWidget(self._pn_buttons['previous starred'], 3, 1, 1, 1)
        self.layout.addWidget(self._pn_buttons['next starred'], 3, 2, 1, 1)
        self.layout.addWidget(self._reset_view_button, 3, 3, 1, 1)
        self.layout.addWidget(self._screenshot_button, 3, 4, 1, 1)

        self.layout.addWidget(self._go_to_id_button, 4, 1, 1, 2)
        self.layout.addWidget(self._id_field, 4, 3, 1, 2)
        self.layout.addWidget(self._go_to_index_button, 5, 1, 1, 2)
        self.layout.addWidget(self._index_field, 5, 3, 1, 2)

    def load_object(self):
        self._id_button.setText('ID {}'.format(self.rd.id))
        self._number_of_obj_label.setText('(#{} / {})'.format(self.rd.j + 1, self.rd.n_objects))
        self._star_button.setIcon(QtGui.QIcon(self.get_star_icon(self.rd.df.at[self.rd.id, 'starred'])))

        self._pn_buttons['previous starred'].setEnabled(np.sum(self.rd.df['starred']) > 0)
        self._pn_buttons['next starred'].setEnabled(np.sum(self.rd.df['starred']) > 0)
        self._dark_mode.setEnabled(False)

    def previous_next_object(self, command: str, starred: bool):
        j_upd = self.update_index(self.rd.j, self.rd.n_objects, command)

        if starred:
            if np.sum(self.rd.df['starred']) > 0:
                while not self.rd.df.iat[j_upd, self.rd.df.columns.get_loc('starred')]:
                    j_upd = self.update_index(j_upd, self.rd.n_objects, command)
            else:
                return

        self.object_selected.emit(j_upd)

    @staticmethod
    def update_index(current_index, n_objects, command: str):
        j_upd = current_index

        if command == 'next':
            j_upd += 1
        elif command == 'previous':
            j_upd -= 1

        j_upd = j_upd % n_objects

        return j_upd

    @staticmethod
    def get_star_icon(starred=False):
        icon_name = 'star.svg' if starred else 'star-empty.svg'
        return get_icon_abs_path(icon_name)

    def star(self):
        starred = not self.rd.df.at[self.rd.id, 'starred']

        self.rd.df.at[self.rd.id, 'starred'] = starred
        self._star_button.setIcon(QtGui.QIcon(self.get_star_icon(starred)))

        self._pn_buttons['previous starred'].setEnabled(np.sum(self.rd.df['starred']) > 0)
        self._pn_buttons['next starred'].setEnabled(np.sum(self.rd.df['starred']) > 0)

    def screenshot(self):
        default_filename = '{}_ID{}.png'.format(self.rd.output_path.stem.replace(' ', '_'), self.rd.id)
        path, extension = qtpy.compat.getsavefilename(self, caption='Save/Save As',
                                                      basedir=str(pathlib.Path().resolve() / default_filename),
                                                      filters='Images (*.png)')

        if path:
            self.screenshot_button_clicked.emit(path)

    def go_to_id(self):
        text = self._id_field.text()
        self._id_field.clear()

        try:
            id_upd = int(text)
        except ValueError:
            logger.error('Invalid ID')
            return

        if id_upd in self.rd.df.index:
            self.object_selected.emit(self.rd.df.index.get_loc(id_upd))
        else:
            logger.error('ID not found')
            return

    def go_to_index(self):
        text = self._index_field.text()
        self._index_field.clear()

        try:
            index_upd = int(text)
        except ValueError:
            logger.error('Invalid index')
            return

        if 0 < index_upd <= self.rd.n_objects:
            self.object_selected.emit(index_upd - 1)
        else:
            logger.error('Index out of range')
            return


def get_icon_abs_path(icon_name: str) -> str:
    return str(pathlib.Path(__file__).parent.parent / 'data' / 'icons' / icon_name)



from dataclasses import dataclass

from ..utils.params import Params


@dataclass
class Cache(Params):
    last_inspection_file: str | None
    last_object_index: int | None
    dock_state: dict | None



import logging
import abc
import pathlib

import numpy as np
from astropy.table import Table
from astropy.io.fits.header import Header

from .AbstractWidget import AbstractWidget

from ..runtime.appdata import AppData
from ..runtime import config
from ..io.viewer_data import get_filename, load
from ..utils import tables


logger = logging.getLogger(__name__)


class ViewerElement(AbstractWidget):
    def __init__(self, rd: AppData, cfg: config.ViewerElement, alias: str, parent=None):
        super().__init__(cfg=cfg, parent=parent)

        self.rd = rd
        self.cfg = cfg
        self.alias: str = alias

        self.filename: pathlib.Path | None = None
        self.data: np.ndarray | Table | None = None
        self.meta: dict | Header | None = None

        self.layout.setContentsMargins(0, 0, 0, 0)

    def load_object(self):
        self.filename = get_filename(self.rd.config.data.dir, self.cfg.filename_keyword, self.rd.id)

        if self.filename is None:
            logger.warning('{} not found (object ID: {})'.format(self.alias, self.rd.id))
            self.data, self.meta = None, None
            return

        if self.cfg.loader_config is None:
            loader_config = {}
        else:
            loader_config = self.cfg.loader_config

        try:
            self.data, self.meta = load(self.cfg.loader, self.filename, self.alias, **loader_config)
        except TypeError as e:
            # unexpected keyword(s) passed to the loader
            logger.error(e.args[0])
            self.data, self.meta = None, None
            return

        if isinstance(self.data, Table):
            # translate the table columns
            if self.rd.config.data.translate:
                tables.translate(self.data, self.rd.config.data.translate)

    @abc.abstractmethod
    def display(self):
        pass

    @abc.abstractmethod
    def reset_view(self):
        pass

    @abc.abstractmethod
    def clear_content(self):
        pass
